//===------------------------------------------------------------*- C++ -*-===//
//
// T-EMU: The Terma Emulator
// (c) Terma 2015
// Authors: Mattias Holm <maho (at) terma.com>
//
//===----------------------------------------------------------------------===//

#ifndef TEMU_MEMORY_IF_H
#define TEMU_MEMORY_IF_H

#include "temu-c/Support/Objsys.h"
#include "temu-c/Support/Memory.h"

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef enum temu_MemoryKind {
  teMK_RAM,
  teMK_ROM,
  teMK_MMIO,
} temu_MemoryKind;

typedef enum {
  teIT_Cpu,
  teIT_Device,
  teIT_Unknown,
} temu_InitiatorType;

// Memory transaction flags
#define TEMU_MT_CACHEABLE         1
#define TEMU_MT_BYPASS_CACHE     (1 << 1)
#define TEMU_MT_FORCE_CACHE_MISS (1 << 2)
#define TEMU_MT_FAILED           (1 << 3)
#define TEMU_MT_CACHE_HIT        (1 << 4)

#define TEMU_MT_PRIV_MASK (7 << 5)
#define TEMU_MT_PRIV_USER (1 << 5)
#define TEMU_MT_PRIV_SUPER (2 << 5)
#define TEMU_MT_PRIV_HYPER (3 << 5)

#define TEMU_MT_PROBE (1 << 8)

//! Generic memory transaction.
//!
//! This type is kept in sync with the emulator core. The layout is guaranteed.
//! and will remain as is, although more fields may be added (at the bottom).
//!
//! When the emulator core issues a memory transaction (assuming no ATC hit),
//! the core allocates this structure on the stack and fills in some of the
//! fields with default values. The memory transaction is passed by pointer or
//! reference. By filling in the different fields you can adapt the result of
//! the memory transaction.
typedef struct temu_MemTransaction {
  uint64_t Va;      //!< 64 bit virtual for unified 32/64 bit interface.
  uint64_t Pa;      //!< 64 bit physical address

  //! Resulting value (or written value). On MMIO Reads the model
  //! fills in this value, and on writes the written value will be
  //! stored here.
  uint64_t Value;

  //! Two-logarithm of the size of the transaction in bytes it is at
  //! most the size of the CPUs max bus size. In case of SPARCv8, this
  //! is 4 bytes (double words are issued as two accesses).  As this
  //! is the 2-log of the size in bytes, a single byte access will
  //! have a size of 0, a 2 byte transaction will have size 1, a 4
  //! byte transaction will have size 2 and an 8 byte transaction will
  //! have size 3.
  uint8_t Size;

  //! Used for device models, this will be filled in with the offset
  //! from the start address of the device (note it is in practice
  //! possible to add a device at multiple locations (which happens in
  //! some rare cases)).
  uint64_t Offset;

  //! InitiatorType identifies the type of object starting the transaction.
  //! this is only relevant when Initiator is set to non-null, and allows
  //! for the specification of device pointers in the initator. This field
  //! is new in TEMU 2.2. The field was introduced to support the
  //! implementation of I/O MMUs.
  temu_InitiatorType InitiatorType;
  //! Initiator of the transaction (a CPU object). It can be null,
  //! which indicate that the transaction was not initiated by normal
  //! CPU activity (e.g. fetch, read or write). When the initiator is
  //! null, models should not attempt to stop the processor core, go
  //! to idle mode or raise traps (posting events is fine). An example
  //! when the initiator is null is when an MMU does a table walk. The
  //! MMU will normally special case handle table walks which access
  //! un-mapped memory.
  void *Initiator;
  //! Page pointer (for caching), this can be filled in by a memory
  //! model to have the emulator core inject the page translation in
  //! the ATC. If a model sets this, the page pointer will
  //! automatically be cleared if the page has attributes
  //! (breakpoints, etc). Models that implement normal memory mapped
  //! registers should NOT populate the page pointer.
  void *Page;
  uint64_t Cycles; //!< Cycle cost for memory access (initialised to 0)

  uint32_t Flags; //!< Flags for use in the memory hierarchy
} temu_MemTransaction;

//! Memory access interface implemented by all memory mapped devices
//! Exposed to the emulator core by a memory object.
typedef struct temu_MemAccessIface {
  //! Function called on fetches. The function can be null in case
  //! fetches are not allowed from the model.
  void (*fetch)(void *Obj, temu_MemTransaction *Mt);

  //! Function called on reads.
  void (*read)(void *Obj, temu_MemTransaction *Mt);

  //! Function called on writes.
  void (*write)(void *Obj, temu_MemTransaction *Mt);

  //! Function called on atomic exchanges, by default if this is not
  //! defined, the memory space will call read followed by write in
  //! order.
  void (*exchange)(void *Obj, temu_MemTransaction *Mt);

  // Optional method, experimental (signature may change)
  void (*mapped)(void *Obj, uint64_t Pa, uint64_t Len);
} temu_MemAccessIface;
#define TEMU_MEM_ACCESS_IFACE_TYPE "MemAccessIface"
TEMU_IFACE_REFERENCE_TYPE(temu_MemAccess);

//! For objects which have actual memory (not just registers) This is
//! for the simulator (not the emu core). The procedures should write
//! the data given in bytes to the given physical offset. The offset
//! is a 64 bit uint to support 64 bit targets. The interface is used
//! for example by DMA transactions.
//!
//! The size argument is in bytes.
//!
//! The swap argument is used to swap bytes to the host
//! endianess. Specify the log size of the read data types.
//!   - 0:  We are reading bytes (bytes will be in target memory order)
//!   - 1:  We are reading half words (will be swapped to host order)
//!   - 2:  We are reading words (will be swapped)
//!   - 3:  We are reading double words (will be swapped)
//! With 0 for swap, we are basically reading a byte array
//!
//! readBytes and writeBytes should return the number of bytes
//! read / written or negative on error.

typedef struct temu_MemoryIface {
  int (*readBytes)(void *Obj, void *Dest, uint64_t Offs, uint32_t Size, int Swap);
  int (*writeBytes)(void *Obj, uint64_t Offs, uint32_t Size, const void *Src, int Swap);
} temu_MemoryIface;
#define TEMU_MEMORY_IFACE_TYPE "MemoryIface"
TEMU_IFACE_REFERENCE_TYPE(temu_Memory);

//! Standard interface for memory space objects
//!
//! Users should not implement their own memory spaces. No stability guarantees
//! are made about this interface at the moment.
typedef struct temu_MemoryMapIface {
  void (*map)(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryKind Type)
     __attribute__((deprecated));
  void (*mapDevice)(void *Obj, uint64_t Addr, uint64_t Len, void *Device)
    __attribute__((deprecated));
  void (*setAttribute)(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr)
    __attribute__((deprecated));
  void (*clearAttribute)(void *Obj, uint64_t Addr, uint64_t Len,
                         temu_MemoryAttr Attr)
    __attribute__((deprecated));
  void (*clearAttributes)(void *Obj, uint64_t Addr, uint64_t Len)
    __attribute__((deprecated));
} temu_MemoryMapIface;

#define TEMU_MEMORY_MAP_IFACE_TYPE "MemoryMapIface"
TEMU_IFACE_REFERENCE_TYPE(temu_MemoryMap);

/*!
 * Map memory object into address space
 * \param Obj The memory space object
 * \param Addr Physical address where to map the device
 * \param Len Length in bytes of area where the object is mapped.
 * \param MemObj The memory object. This object must correspond to the
 *               MemAccessIface
 * \result Zero on success, other values indicates that the mapping failed
 */
int
temu_mapMemorySpace(void *Obj, uint64_t Addr, uint64_t Len, void *MemObj);


/*!
 * Map memory object into address space with flags
 * \param Obj The memory space object
 * \param Addr Physical address where to map the device
 * \param Len Length in bytes of area where the object is mapped.
 * \param MemObj The memory object. This object must correspond to the
 *               MemAccessIface
 * \param Flags Sticky flags for memory accesses to that object (e.g. cachable)
 * \result Zero on success, other values indicates that the mapping failed
 */
int
temu_mapMemorySpaceFlags(void *Obj, uint64_t Addr, uint64_t Len, void *MemObj,
                         uint32_t Flags);



/*!
 * Set attribute on memory space location
 * \param Obj The memory space object
 * \param Addr Physical address where to map the device
 * \param Len Length in bytes of area where the attribute should be set.
 * \param Attr The attribute to set.
 */

void
temu_setMemAttr(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr);


/*!
 * Clear attribute on memory space location
 * \param Obj The memory space object
 * \param Addr Physical address where to map the device
 * \param Len Length in bytes of area where the attribute should be set.
 * \param Attr The attribute to clear.
 */
void
temu_clearMemAttr(void *Obj, uint64_t Addr, uint64_t Len, temu_MemoryAttr Attr);

#ifdef __cplusplus
}
#endif

#endif /* ! TEMU_MEMORY_IF_H */
